/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use std::sync::atomic::{AtomicU16, Ordering};
use std::sync::{Arc, OnceLock};

use chrono::{offset::Local, DateTime};
use color_eyre::eyre::{eyre, Result};
use eyeball_im::VectorDiff;
use futures::{FutureExt, Stream, StreamExt};
use itertools::Itertools;
use matrix_sdk::async_trait;
use matrix_sdk::ruma::{OwnedRoomId, RoomId};
use matrix_sdk::Client;
use matrix_sdk::Room;
use matrix_sdk_ui::timeline::{
  BackPaginationStatus, PaginationOptions, RoomExt, Timeline, TimelineItem, TimelineItemKind,
  VirtualTimelineItem,
};
use ratatui::text::Text;
use smallvec::SmallVec;

use super::{Buffer, BufferItem, BufferItemContent};
use crate::widgets::Prerender;

/// Like [`BufferItemContent`] but owned.
#[derive(Debug, Clone)]
pub enum OwnedBufferItemContent {
  Text(String),
  Divider(String),
  Empty,
}

/// Like `format!()` but returns OwnedBufferItemContent::Text
macro_rules! text {
  ($($tokens:tt)*) => {
    OwnedBufferItemContent::Text(format!($($tokens)*))
  }
}

pub struct SingleClientRoomBuffer {
  room_id: OwnedRoomId,
  client: Client,
  items: imbl::vector::Vector<(OwnedBufferItemContent, Prerender)>,
  // TODO: get rid of this trait object, we know it's matrix_sdk_ui::timeline::TimelineStream
  stream: Box<dyn Stream<Item = Vec<VectorDiff<Arc<TimelineItem>>>> + Send + Sync + Unpin>,
  timeline: Arc<Timeline>,
  back_pagination_request: AtomicU16,
}

impl SingleClientRoomBuffer {
  async fn poll_updates(&mut self) {
    let back_pagination_request = self.back_pagination_request.swap(0, Ordering::Relaxed);
    if back_pagination_request > 0 {
      // TODO: run this concurrently with stream.next() below
      self.spawn_back_pagination(back_pagination_request).await;
    }
    if let Some(changes) = self.stream.next().await {
      for change in changes {
        change
          .map(|item| (self.format_timeline_item(item), Prerender::new()))
          .apply(&mut self.items);
      }
    }
  }

  fn format_timeline_item(&self, tl_item: impl AsRef<TimelineItem>) -> OwnedBufferItemContent {
    match tl_item.as_ref().kind() {
      TimelineItemKind::Event(event) => {
        use matrix_sdk_ui::timeline::TimelineItemContent::*;
        let sender = event
          .sender()
          .as_str()
          .strip_prefix("@")
          .expect("missing @ prefix");
        match event.content() {
          Message(message) => text!("   <{}> {}", sender, message.body().replace("\n", "\n    ")),
          RedactedMessage => text!("xx <{}> [redacted]", sender),
          Sticker(sticker) => text!("st <{}> {}", sender, sticker.content().body),
          UnableToDecrypt(_) => text!("xx <{}> [unable to decrypt]", sender),
          MembershipChange(change) => {
            use matrix_sdk_ui::timeline::MembershipChange::*;
            if change.user_id() == event.sender() {
              let Some(change_kind) = change.change() else {
                return text!("--- {} made incomprehensible changes to themselves", sender);
              };
              match change_kind {
                None => text!("--- {} made no discernable changes to themselves", sender),
                Error => text!(
                  "xxx {} made a change to themselves that made matrix-sdk-ui error",
                  sender
                ),
                Joined => text!("--> {} joined", sender),
                Left => text!("<-- {} left", sender),
                Banned => text!("-x- {} banned themselves", sender),
                Unbanned => text!("-x- {} unbanned themselves", sender),
                Kicked => text!("<!- {} kicked themselves", sender),
                Invited => text!("-o- {} invited themselves", sender),
                KickedAndBanned => text!("<!x {} kicked and banned themselves", sender),
                InvitationAccepted => text!("-o> {} accepted an invite", sender),
                InvitationRejected => text!("-ox {} rejected an invite", sender),
                InvitationRevoked => text!("--x {} revoked an invite", sender),
                Knocked => text!("-?> {} knocked", sender),
                KnockAccepted => text!("-?o {} accepted a knock", sender),
                KnockRetracted => text!("-?x {} retracted a knock", sender),
                KnockDenied => text!("-?x {} denied a knock", sender),
                NotImplemented => text!(
                  "xxx {} made a change matrix-sdk-ui does not support yet",
                  sender
                ),
              }
            } else if change.user_id() == "" {
              let Some(change_kind) = change.change() else {
                return text!("--- {} made incomprehensible changes", sender);
              };
              match change_kind {
                None => text!("--- {} made no discernable changes", sender),
                Error => text!("xxx {} made a change that made matrix-sdk-ui error", sender),
                Joined | Left | Banned | Unbanned | Kicked | Invited | KickedAndBanned
                | InvitationAccepted | InvitationRejected | InvitationRevoked | Knocked
                | KnockAccepted | KnockRetracted | KnockDenied => {
                  text!("--> {} made a non-sensical change: {:?}", sender, change)
                },
                NotImplemented => text!(
                  "xxx {} made a change matrix-sdk-ui does not support yet",
                  sender
                ),
              }
            } else {
              let target = change
                .user_id()
                .as_str()
                .strip_prefix("@")
                .expect("missing @ prefix");
              let Some(change_kind) = change.change() else {
                return text!("--- {} made incomprehensible changes to {}", sender, target);
              };
              match change_kind {
                None => text!("--- {} made no discernable changes to {}", sender, target),
                Error => text!(
                  "xxx {} made a change to {} that made matrix-sdk-ui error",
                  sender,
                  target
                ),
                Joined | Left => text!(
                  "--> {} made a non-sensical change to {}: {:?}",
                  sender,
                  target,
                  change
                ),
                Banned => text!("-x- {} banned {}", sender, target),
                Unbanned => text!("-x- {} unbanned {}", sender, target),
                Kicked => text!("<!- {} kicked {}", sender, target),
                Invited => text!("-o- {} invited {}", sender, target),
                KickedAndBanned => text!("<!x {} kicked and banned {}", sender, target),
                InvitationAccepted => text!("-o> {} accepted an invite to {}", sender, target),
                InvitationRejected => text!("-ox {} rejected an invite to {}", sender, target),
                InvitationRevoked => text!("--x {} revoked an invite to {}", sender, target),
                Knocked => text!("-?> {} made {} knock", sender, target),
                KnockAccepted => text!("-?o {} accepted {}'s knock", sender, target),
                KnockRetracted => text!("-?x {} retracted {}'s knock", sender, target),
                KnockDenied => text!("-?x {} denied {}'s knock", sender, target),
                NotImplemented => text!(
                  "xxx {} made a change to {} that matrix-sdk-ui does not support yet",
                  sender,
                  target
                ),
              }
            }
          },

          ProfileChange(_) => text!("--- {} updated their profile", sender),
          OtherState(state) => {
            if state.state_key() == "" {
              text!("--- {} changed the room: {:?}", sender, state.content())
            } else {
              text!(
                "--- {} changed {}: {:?}",
                sender,
                state.state_key(),
                state.content()
              )
            }
          },
          FailedToParseMessageLike { event_type, error } => text!(
            "xxx {} sent a {} message that made matrix-sdk-ui error: {:?}",
            sender,
            event_type,
            error
          ),
          FailedToParseState {
            event_type,
            state_key,
            error,
          } => {
            text!(
              "xxx {} made a {} change to {} that made matrix-sdk-ui error: {:?}",
              sender,
              event_type,
              state_key,
              error
            )
          },
          Poll(_) => text!("-?- {} acted on a poll", sender),
        }
      },
      TimelineItemKind::Virtual(VirtualTimelineItem::ReadMarker) => {
        OwnedBufferItemContent::Divider(format!("read marker"))
      },
      TimelineItemKind::Virtual(VirtualTimelineItem::DayDivider(day_divider)) => {
        match day_divider.to_system_time() {
          Some(system_time) => {
            let datetime: DateTime<Local> = system_time.into();
            OwnedBufferItemContent::Divider(format!("{}", datetime.format("%Y-%m-%d")))
          },
          None => {
            tracing::warn!("Could not convert {:?} to system time", day_divider);
            OwnedBufferItemContent::Empty
          },
        }
      },
    }
  }

  async fn spawn_back_pagination(&self, num: u16) {
    let room_id = self.room_id.clone();
    let timeline = self.timeline.clone();
    let mut back_pagination_status = timeline.back_pagination_status();
    match back_pagination_status.get() {
      BackPaginationStatus::Paginating | BackPaginationStatus::TimelineStartReached => {
        // We are already waiting for a backfill from the server
      },
      BackPaginationStatus::Idle => {
        tokio::spawn(async move {
          tracing::info!("Starting pagination for {}", room_id);
          timeline
            .paginate_backwards(matrix_sdk_ui::timeline::PaginationOptions::until_num_items(
              num, num,
            ))
            .await
            .unwrap_or_else(|e| tracing::error!("Failed to paginate {} backward: {}", room_id, e));
          tracing::info!("Ended pagination for {}", room_id);
        });

        // Wait for the task we just spawned to change the status, so we don't risk starting
        // a new one in the meantime
        back_pagination_status.next().await;
      },
    }
  }
}

pub struct RoomBuffer {
  room_id: OwnedRoomId,

  // It's unlikely users will join the same room with more than one account;
  // avoid a useless heap allocation for the usual case.
  buffers: SmallVec<[SingleClientRoomBuffer; 1]>,
}

fn f(b: RoomBuffer) {
  g(b);
}
fn g<B: Send>(b: B) {}

impl RoomBuffer {
  pub async fn new(initial_client: Client, room_id: OwnedRoomId) -> Result<Self> {
    let mut self_ = RoomBuffer {
      buffers: SmallVec::new(),
      room_id,
    };
    self_.add_client(initial_client).await?;
    Ok(self_)
  }

  pub async fn add_client(&mut self, client: Client) -> Result<()> {
    let timeline = client
      .get_room(&self.room_id)
      .ok_or_else(|| {
        tracing::error!(
          "Adding {:?} for {:?}, but it does not know this room ({} other clients know this room)",
          client,
          self.room_id,
          self.buffers.len()
        );
        eyre!("Unknown room {} for client {:?}", self.room_id, client)
      })?
      .timeline_builder()
      .build()
      .await;
    let (items, stream) = timeline.subscribe_batched().await;
    tracing::info!(
      "Added client for {}, initial items: {:?}",
      self.room_id,
      items
    );
    self.buffers.push(SingleClientRoomBuffer {
      room_id: self.room_id.clone(),
      client,
      timeline: Arc::new(timeline),
      items: items // FIXME: it's always empty. why?
        .into_iter()
        .map(|item| (text!("Initial item: {:#?}", item), Prerender::new()))
        .collect(),
      stream: Box::new(stream),
      back_pagination_request: AtomicU16::new(0),
    });
    Ok(())
  }
}

#[async_trait]
impl Buffer for RoomBuffer {
  fn short_name(&self) -> String {
    self
      .buffers
      .iter()
      .flat_map(|buf| buf.client.get_room(&self.room_id))
      .flat_map(|room| room.canonical_alias()) // TODO: .display_name() is better, but async :(
      .map(|alias| alias.as_str().to_owned())
      .next()
      .unwrap_or(self.room_id.as_str().to_owned())
  }

  fn room_id(&self) -> Option<&RoomId> {
    Some(&self.room_id)
  }

  async fn poll_updates(&mut self) {
    futures::future::join_all(
      self
        .buffers
        .iter_mut()
        .map(SingleClientRoomBuffer::poll_updates),
    )
    .await;
  }

  fn content<'a>(&'a self) -> Box<dyn Iterator<Item = BufferItem<'a>> + 'a> {
    // TODO: merge buffers, etc.
    Box::new(
      self
        .buffers
        .first()
        .unwrap_or_else(|| panic!("No sub-buffer for {}", self.room_id))
        .items
        .iter()
        .rev()
        .map(|(line, prerender)| BufferItem {
          content: match line {
            OwnedBufferItemContent::Text(text) => BufferItemContent::Text(Text::raw(text)),
            OwnedBufferItemContent::Divider(text) => BufferItemContent::Divider(Text::raw(text)),
            OwnedBufferItemContent::Empty => BufferItemContent::Empty,
          },
          prerender,
        }),
    )
  }

  fn request_back_pagination(&self, num: u16) {
    // TODO: pick a client at random instead of just the first one, etc.
    self
      .buffers
      .first()
      .unwrap_or_else(|| panic!("No sub-buffer for {}", self.room_id))
      .back_pagination_request
      .fetch_max(num, Ordering::Relaxed);
  }
}
